// ApertureRoi.java

package astroj;

import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import java.awt.geom.*;

/**
 * A ROI consisting of three concentric circles (inner for object, outer annulus for background).
 * The ROI keeps track both of it's double position and a boolean flag which determines whether
 * the backround aperture should be displayed or not.
 */
public class ApertureRoi extends Roi
	{
	protected String valueText = null;
    protected String nameText = null;
	protected Font font = null;
	protected double xPos,yPos,srcMinusBack;
	protected double r1,r2,r3;
    protected Color apColor;

	protected boolean showAperture = true;
	protected boolean showSky = false;
    protected boolean showName = false;
	protected boolean showValue = false;
    AstroCanvas ac = null;
    AffineTransform canvTrans = null;
    boolean netFlipX=false, netFlipY=false, netRotate=false;

	public ApertureRoi (double x, double y, double rad1, double rad2, double rad3, double src)
		{
		super ((int)x,(int)y,1,1);

		xPos = x;		// Roi CLASS HAS PRIVATE FLOATING POINT COORDINATES
		yPos = y;
		r1 = rad1;
		r2 = rad2;
		r3 = rad3;
		srcMinusBack = src;
		font = new Font ("SansSerif", Font.PLAIN, 16);
		}
    
    public void setSize(double rad1, double rad2, double rad3)
        {
		r1 = rad1;
		r2 = rad2;
		r3 = rad3;
        }    

	/**
	 * Sets the appearance of the ROI when it is displayed.
	 */
	public void setAppearance (boolean aper, boolean sky, boolean name, boolean label, Color col, String ntext, String vtext)
		{
		showAperture = aper;
		showSky = sky;
        showName = name;
		showValue = label;
        apColor = col;
		valueText = vtext;
        nameText = ntext;
		}
    public void setShowName(boolean name)
        {
        showName = name;
        }
    public void setShowValue(boolean label)
        {
        showValue = label;
        }
    public void setShowSky(boolean sky)
        {
        showSky = sky;
        }
    public void setApColor(Color col)
        {
        apColor = col;
        }

    public double getXpos()
        {
        return xPos;
        }
    public double getYpos()
        {
        return yPos;
        }
    public double getRadius()
        {
        return r1;
        }
    public boolean getShowName()
        {
        return showAperture;
        }
    public boolean getShowValue()
        {
        return showValue;
        }
    public boolean getShowSky()
        {
        return showSky;
        }
    public Color getApColor()
        {
        return apColor;
        }



	/**
	 * Lists the ROI contents.
	 */
	void log()
		{
		IJ.log("ApertureRoi: x,y="+xPos+","+yPos+", src-back="+srcMinusBack+", r1="+r1+", r2="+r2+", r3="+r3);
		}

	/**
	 * Returns an array with the double position.
	 */
	public double[] getCenter()
		{
		double[] d = new double[] {xPos,yPos};
		return d;
		}

	/**
	 * Returns the measurement value associated with the ROI.
	 */
	public double getMeasurement()
		{
		return srcMinusBack;
		}

	/**
	 * Returns true if the screen position (xs,ys) is within the currently displayed area of the ROI
	 * (i.e. inner circle if showSky==false).
	 */
	public boolean contains (int xs, int ys)
		{
		int xx,yy,ww,hh;
		if (showSky)
			{
			xx = (int)(xPos-r3);
			ww = (int)(xPos+r3)-xx;
			yy = (int)(yPos-r3);
			hh = (int)(yPos+r3)-yy;
			}
		else	{
			xx = (int)(xPos-r1);
			ww = (int)(xPos+r1)-xx;
			yy = (int)(yPos-r1);
			hh = (int)(yPos+r1)-yy;
			}
		if (xs < xx || xs > xx+ww) return false;
		if (ys < yy || ys > yy+hh) return false;
		return true;
		}

	/**
	 * Displays the aperture either as a simple circle (showSky false) or as three circles (showSky true).
	 */
	public void draw (Graphics g)
		{
        boolean aij = false;
        if (ic instanceof astroj.AstroCanvas)
            {
            aij =  true;
            ac =(AstroCanvas)ic;
            canvTrans = ((Graphics2D)g).getTransform();
            ((Graphics2D)g).setTransform(ac.invCanvTrans);
            netFlipX = ac.getNetFlipX();
            netFlipY = ac.getNetFlipY();
            netRotate = ac.getNetRotate();
            }
		g.setColor (apColor);
		FontMetrics metrics = g.getFontMetrics (font);
		int h = metrics.getHeight();
		int w = metrics.stringWidth(valueText)+3;
		int descent = metrics.getDescent();
		g.setFont (font);

		int sx = screenXD (xPos);
		int sy = screenYD (yPos);
		int x1 = netFlipX ? screenXD (xPos+r1) : screenXD (xPos-r1);
		int w1 = netFlipX ? screenXD (xPos-r1)-x1 : screenXD (xPos+r1)-x1;
		int y1 = netFlipY ? screenYD (yPos+r1) : screenYD (yPos-r1);
		int h1 = netFlipY ? screenYD (yPos-r1)-y1 : screenYD (yPos+r1)-y1;

//        if (aij) ac.transEnabled = true;
		int xl = sx + h;
		int yl = sy + h/3;
//        if (aij) ac.transEnabled = false;

		if (showAperture)
			{
			g.drawOval (x1,y1,w1,h1);
//            IJ.log("sx="+sx+"  sy="+sy+"  x1="+x1+"  y1="+y1+"  w1="+w1+"  h1="+h1);
//            if (aij) ac.transEnabled = true;
//            int xl1 = netFlipX ? screenXD (xPos+r1) : screenXD (xPos-r1);
//            int wl1 = netFlipX ? screenXD (xPos-r1)-xl1 : screenXD (xPos+r1)-xl1;
//            if (aij) ac.transEnabled = false;
			xl = x1+w1+descent;
			}

		if (showSky)
			{
			int x2 = netFlipX ? screenXD (xPos+r2) : screenXD (xPos-r2);
			int x3 = netFlipX ? screenXD (xPos+r3) : screenXD (xPos-r3);
			int w2 = netFlipX ? screenXD (xPos-r2)-x2 : screenXD (xPos+r2)-x2;
			int w3 = netFlipX ? screenXD (xPos-r3)-x3 : screenXD (xPos+r3)-x3;
			int y2 = netFlipY ? screenYD (yPos+r2) : screenYD (yPos-r2);
			int y3 = netFlipY ? screenYD (yPos+r3) : screenYD (yPos-r3);
			int h2 = netFlipY ? screenYD (yPos-r2)-y2 : screenYD (yPos+r2)-y2;
			int h3 = netFlipY ? screenYD (yPos-r3)-y3 : screenYD (yPos+r3)-y3;
			g.drawOval (x2,y2,w2,h2);
			g.drawOval (x3,y3,w3,h3);
//            if (aij) ac.transEnabled = true;
//            int xl3 = netFlipX ? screenXD (xPos+r3) : screenXD (xPos-r3);
//            int wl3 = netFlipX ? screenXD (xPos-r3)-xl3 : screenXD (xPos+r3)-xl3;
//            if (aij) ac.transEnabled = false;
			xl = x3+w3+descent;
			}

//        if (aij) ((Graphics2D)g).setTransform(ac.invCanvTrans);

		if (showName && showValue && nameText != null && !nameText.equals("") && valueText != null && !valueText.equals(""))
			g.drawString (nameText+"="+valueText, xl,yl);
        else if(showName && nameText != null && !nameText.equals(""))
			g.drawString (nameText, xl,yl);
        else if(showValue && valueText != null && !valueText.equals(""))
			g.drawString (valueText, xl,yl);
        if (aij) ((Graphics2D)g).setTransform(canvTrans);
		}

	/**
	 * Calculates the CORRECT screen x-position of a given decimal pixel position.
	 */
//	public int fscreenX (double x) { return netFlipX ? (int)(ac.getWidth()-ac.screenXD(x)-1): ac.screenXD(x); } //-((ac.getWidth())%ac.getMagnification())
//
//	/**
//	 * Calculates the CORRECT screen y-position of a given decimal pixel position.
//	 */
//	public int fscreenY (double y) { return netFlipY ? (int)(ac.getHeight()-ac.screenYD(y)-1): ac.screenYD(y); } //-(ac.getHeight()%ac.getMagnification())
    
	}

